"""
This file contains coordinate mappings between physical coordinates and those
defined on unit elements, as well as doing the corresponding intracell
interpolation on finite element data.


"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

cimport numpy as np
from numpy cimport ndarray
cimport cython
import numpy as np
from libc.math cimport fabs
from yt.utilities.lib.autogenerated_element_samplers cimport \
    Q1Function3D, \
    Q1Jacobian3D, \
    Q1Function2D, \
    Q1Jacobian2D, \
    Q2Function2D, \
    Q2Jacobian2D, \
    W1Function3D, \
    W1Jacobian3D, \
    T2Function2D, \
    T2Jacobian2D, \
    Tet2Function3D, \
    Tet2Jacobian3D

cdef extern from "platform_dep.h":
    double fmax(double x, double y) nogil

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double determinant_3x3(double* col0,
                            double* col1,
                            double* col2) nogil:
    return col0[0]*col1[1]*col2[2] - col0[0]*col1[2]*col2[1] - \
           col0[1]*col1[0]*col2[2] + col0[1]*col1[2]*col2[0] + \
           col0[2]*col1[0]*col2[1] - col0[2]*col1[1]*col2[0]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double maxnorm(double* f, int dim) nogil:
    cdef double err
    cdef int i
    err = fabs(f[0])
    for i in range(1, dim):
        err = fmax(err, fabs(f[i]))
    return err


cdef class ElementSampler:
    '''

    This is a base class for sampling the value of a finite element solution
    at an arbitrary point inside a mesh element. In general, this will be done
    by transforming the requested physical coordinate into a mapped coordinate
    system, sampling the solution in mapped coordinates, and returning the result.
    This is not to be used directly; use one of the subclasses instead.

    '''

    def __init__(self):
        self.inclusion_tol = 1.0e-8

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self,
                               double* mapped_x,
                               double* vertices,
                               double* physical_x) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_real_point(self,
                                     double* vertices,
                                     double* field_values,
                                     double* physical_x) nogil:
        cdef double val
        cdef double mapped_coord[4]

        self.map_real_to_unit(mapped_coord, vertices, physical_x)
        val = self.sample_at_unit_point(mapped_coord, field_values)
        return val


cdef class P1Sampler1D(ElementSampler):
    '''

    This implements sampling inside a linear, 1D element.

    '''

    def __init__(self):
        super(P1Sampler1D, self).__init__()
        self.num_mapped_coords = 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self, double* mapped_x,
                               double* vertices, double* physical_x) nogil:
        mapped_x[0] = -1.0 + 2.0*(physical_x[0] - vertices[0]) / (vertices[1] - vertices[0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        return vals[0] * (1 - coord[0]) / 2.0 + vals[1] * (1.0 + coord[0]) / 2.0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol):
            return 0
        return 1


cdef class P1Sampler2D(ElementSampler):
    '''

    This implements sampling inside a linear, triangular mesh element.
    This mapping is linear and can be inverted easily. Note that this
    implementation uses triangular (or barycentric) coordinates.

    '''

    def __init__(self):
        super(P1Sampler2D, self).__init__()
        self.num_mapped_coords = 3


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self, double* mapped_x,
                               double* vertices, double* physical_x) nogil:

        cdef double[3] col0
        cdef double[3] col1
        cdef double[3] col2

        col0[0] = vertices[0]
        col0[1] = vertices[1]
        col0[2] = 1.0

        col1[0] = vertices[2]
        col1[1] = vertices[3]
        col1[2] = 1.0

        col2[0] = vertices[4]
        col2[1] = vertices[5]
        col2[2] = 1.0

        det = determinant_3x3(col0, col1, col2)

        mapped_x[0] = ((vertices[3] - vertices[5])*physical_x[0] + \
                       (vertices[4] - vertices[2])*physical_x[1] + \
                       (vertices[2]*vertices[5] - vertices[4]*vertices[3])) / det

        mapped_x[1] = ((vertices[5] - vertices[1])*physical_x[0] + \
                       (vertices[0] - vertices[4])*physical_x[1] + \
                       (vertices[4]*vertices[1] - vertices[0]*vertices[5])) / det

        mapped_x[2] = 1.0 - mapped_x[1] - mapped_x[0]


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        return vals[0]*coord[0] + vals[1]*coord[1] + vals[2]*coord[2]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for triangles, we check whether all mapped_coords are
        # between 0 and 1, to within the inclusion tolerance
        cdef int i
        for i in range(3):
            if (mapped_coord[i] < -self.inclusion_tol or
                mapped_coord[i] - 1.0 > self.inclusion_tol):
                return 0
        return 1


cdef class P1Sampler3D(ElementSampler):
    '''

    This implements sampling inside a linear, tetrahedral mesh element.
    This mapping is linear and can be inverted easily.

    '''


    def __init__(self):
        super(P1Sampler3D, self).__init__()
        self.num_mapped_coords = 4

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self, double* mapped_x,
                               double* vertices, double* physical_x) nogil:

        cdef int i
        cdef double d
        cdef double[3] bvec
        cdef double[3] col0
        cdef double[3] col1
        cdef double[3] col2

        # here, we express positions relative to the 4th element,
        # which is selected by vertices[9]
        for i in range(3):
            bvec[i] = physical_x[i]       - vertices[9 + i]
            col0[i] = vertices[0 + i]     - vertices[9 + i]
            col1[i] = vertices[3 + i]     - vertices[9 + i]
            col2[i] = vertices[6 + i]     - vertices[9 + i]

        d = determinant_3x3(col0, col1, col2)
        mapped_x[0] = determinant_3x3(bvec, col1, col2)/d
        mapped_x[1] = determinant_3x3(col0, bvec, col2)/d
        mapped_x[2] = determinant_3x3(col0, col1, bvec)/d
        mapped_x[3] = 1.0 - mapped_x[0] - mapped_x[1] - mapped_x[2]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        return vals[0]*coord[0] + vals[1]*coord[1] + \
            vals[2]*coord[2] + vals[3]*coord[3]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for tetrahedra, we check whether all mapped coordinates
        # are within 0 and 1, to within the inclusion tolerance
        cdef int i
        for i in range(4):
            if (mapped_coord[i] < -self.inclusion_tol or
                mapped_coord[i] - 1.0 > self.inclusion_tol):
                return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        cdef double u, v, w
        cdef double thresh = 2.0e-2
        if mapped_coord[0] == 0:
            u = mapped_coord[1]
            v = mapped_coord[2]
            w = mapped_coord[3]
        elif mapped_coord[1] == 0:
            u = mapped_coord[2]
            v = mapped_coord[3]
            w = mapped_coord[0]
        elif mapped_coord[2] == 0:
            u = mapped_coord[1]
            v = mapped_coord[3]
            w = mapped_coord[0]
        else:
            u = mapped_coord[1]
            v = mapped_coord[2]
            w = mapped_coord[0]
        if ((u < thresh) or
            (v < thresh) or
            (w < thresh) or
            (fabs(u - 1) < thresh) or
            (fabs(v - 1) < thresh) or
            (fabs(w - 1) < thresh)):
            return 1
        return -1


cdef class NonlinearSolveSampler3D(ElementSampler):

    '''

    This is a base class for handling element samplers that require
    a nonlinear solve to invert the mapping between coordinate systems.
    To do this, we perform Newton-Raphson iteration using a specified
    system of equations with an analytic Jacobian matrix. This solver
    is hard-coded for 3D for reasons of efficiency. This is not to be
    used directly, use one of the subclasses instead.

    '''

    def __init__(self):
        super(NonlinearSolveSampler3D, self).__init__()
        self.tolerance = 1.0e-9
        self.max_iter = 10

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self,
                               double* mapped_x,
                               double* vertices,
                               double* physical_x) nogil:
        '''

        A thorough description of Newton's method and modifications for global
        convergence can be found in Dennis's text "Numerical Methods for
        Unconstrained Optimization and Nonlinear Equations."

        x: solution vector; holds unit/mapped coordinates
        xk: temporary vector for holding solution of current iteration
        f: residual vector
        r, s, t: three columns of Jacobian matrix corresponding to unit/mapped
                 coordinates r, s, and t
        d: Jacobian determinant
        s_n: Newton step vector
        lam: fraction of Newton step by which to change x
        alpha: constant proportional to how much residual required to decrease.
               1e-4 is value of alpha recommended by Dennis
        err_c: Error of current iteration
        err_plus: Error of next iteration
        min_lam: minimum fraction of Newton step that the line search is allowed
                 to take. General experience suggests that lambda values smaller
                 than 1e-3 will not significantly reduce the residual, but we
                 set to 1e-6 just to be safe
        '''
        
        cdef int i
        cdef double d, lam
        cdef double[3] f
        cdef double[3] r
        cdef double[3] s
        cdef double[3] t
        cdef double[3] x, xk, s_n
        cdef int iterations = 0
        cdef double err_c, err_plus
        cdef double alpha = 1e-4
        cdef double min_lam = 1e-6

        # initial guess
        for i in range(3):
            x[i] = 0.0

        # initial error norm
        self.func(f, x, vertices, physical_x)
        err_c = maxnorm(f, 3)

        # begin Newton iteration
        while (err_c > self.tolerance and iterations < self.max_iter):
            self.jac(r, s, t, x, vertices, physical_x)
            d = determinant_3x3(r, s, t)

            s_n[0] = - (determinant_3x3(f, s, t)/d)
            s_n[1] = - (determinant_3x3(r, f, t)/d)
            s_n[2] = - (determinant_3x3(r, s, f)/d)
            xk[0] = x[0] + s_n[0]
            xk[1] = x[1] + s_n[1]
            xk[2] = x[2] + s_n[2]
            self.func(f, xk, vertices, physical_x)
            err_plus = maxnorm(f, 3)
            
            lam = 1
            while err_plus > err_c * (1. - alpha * lam) and lam > min_lam:
                lam = lam / 2
                xk[0] = x[0] + lam * s_n[0]
                xk[1] = x[1] + lam * s_n[1]
                xk[2] = x[2] + lam * s_n[2]
                self.func(f, xk, vertices, physical_x)
                err_plus = maxnorm(f, 3)

            x[0] = xk[0]
            x[1] = xk[1]
            x[2] = xk[2]
            err_c = err_plus
            iterations += 1

        if (err_c > self.tolerance):
            # we did not converge, set bogus value
            for i in range(3):
                mapped_x[i] = -99.0
        else:
            for i in range(3):
                mapped_x[i] = x[i]


cdef class Q1Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, linear, hexahedral mesh element.

    '''

    def __init__(self):
        super(Q1Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = Q1Function3D
        self.jac = Q1Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, rm, rp, sm, sp, tm, tp

        rm = 1.0 - coord[0]
        rp = 1.0 + coord[0]
        sm = 1.0 - coord[1]
        sp = 1.0 + coord[1]
        tm = 1.0 - coord[2]
        tp = 1.0 + coord[2]

        F = vals[0]*rm*sm*tm + vals[1]*rp*sm*tm + vals[2]*rp*sp*tm + vals[3]*rm*sp*tm + \
            vals[4]*rm*sm*tp + vals[5]*rp*sm*tp + vals[6]*rp*sp*tp + vals[7]*rm*sp*tp
        return 0.125*F

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for hexes, the mapped coordinates all go from -1 to 1
        # if we are inside the element.
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[2]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        if (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
            fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        else:
            return -1


cdef class S2Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, 20-node hexahedral mesh element.

    '''

    def __init__(self):
        super(S2Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = S2Function3D
        self.jac = S2Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, r, s, t, rm, rp, sm, sp, tm, tp

        r = coord[0]
        rm = 1.0 - r
        rp = 1.0 + r

        s = coord[1]
        sm = 1.0 - s
        sp = 1.0 + s

        t = coord[2]
        tm = 1.0 - t
        tp = 1.0 + t

        F = rm*sm*tm*(-r - s - t - 2.0)*vals[0] \
          + rp*sm*tm*( r - s - t - 2.0)*vals[1] \
          + rp*sp*tm*( r + s - t - 2.0)*vals[2] \
          + rm*sp*tm*(-r + s - t - 2.0)*vals[3] \
          + rm*sm*tp*(-r - s + t - 2.0)*vals[4] \
          + rp*sm*tp*( r - s + t - 2.0)*vals[5] \
          + rp*sp*tp*( r + s + t - 2.0)*vals[6] \
          + rm*sp*tp*(-r + s + t - 2.0)*vals[7] \
          + 2.0*(1.0 - r*r)*sm*tm*vals[8]  \
          + 2.0*rp*(1.0 - s*s)*tm*vals[9]  \
          + 2.0*(1.0 - r*r)*sp*tm*vals[10] \
          + 2.0*rm*(1.0 - s*s)*tm*vals[11] \
          + 2.0*rm*sm*(1.0 - t*t)*vals[12] \
          + 2.0*rp*sm*(1.0 - t*t)*vals[13] \
          + 2.0*rp*sp*(1.0 - t*t)*vals[14] \
          + 2.0*rm*sp*(1.0 - t*t)*vals[15] \
          + 2.0*(1.0 - r*r)*sm*tp*vals[16] \
          + 2.0*rp*(1.0 - s*s)*tp*vals[17] \
          + 2.0*(1.0 - r*r)*sp*tp*vals[18] \
          + 2.0*rm*(1.0 - s*s)*tp*vals[19]
        return 0.125*F

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[2]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        if (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
            fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        else:
            return -1


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void S2Function3D(double* fx,
                              double* x,
                              double* vertices,
                              double* phys_x) nogil:
        cdef int i
        cdef double r, s, t, rm, rp, sm, sp, tm, tp

        r = x[0]
        rm = 1.0 - r
        rp = 1.0 + r

        s = x[1]
        sm = 1.0 - s
        sp = 1.0 + s

        t = x[2]
        tm = 1.0 - t
        tp = 1.0 + t

        for i in range(3):
            fx[i] = rm*sm*tm*(-r - s - t - 2.0)*vertices[0 + i]  \
                  + rp*sm*tm*( r - s - t - 2.0)*vertices[3 + i]  \
                  + rp*sp*tm*( r + s - t - 2.0)*vertices[6 + i]  \
                  + rm*sp*tm*(-r + s - t - 2.0)*vertices[9 + i]  \
                  + rm*sm*tp*(-r - s + t - 2.0)*vertices[12 + i] \
                  + rp*sm*tp*( r - s + t - 2.0)*vertices[15 + i] \
                  + rp*sp*tp*( r + s + t - 2.0)*vertices[18 + i] \
                  + rm*sp*tp*(-r + s + t - 2.0)*vertices[21 + i] \
                  + 2.0*(1.0 - r*r)*sm*tm*vertices[24 + i] \
                  + 2.0*rp*(1.0 - s*s)*tm*vertices[27 + i] \
                  + 2.0*(1.0 - r*r)*sp*tm*vertices[30 + i] \
                  + 2.0*rm*(1.0 - s*s)*tm*vertices[33 + i] \
                  + 2.0*rm*sm*(1.0 - t*t)*vertices[36 + i] \
                  + 2.0*rp*sm*(1.0 - t*t)*vertices[39 + i] \
                  + 2.0*rp*sp*(1.0 - t*t)*vertices[42 + i] \
                  + 2.0*rm*sp*(1.0 - t*t)*vertices[45 + i] \
                  + 2.0*(1.0 - r*r)*sm*tp*vertices[48 + i] \
                  + 2.0*rp*(1.0 - s*s)*tp*vertices[51 + i] \
                  + 2.0*(1.0 - r*r)*sp*tp*vertices[54 + i] \
                  + 2.0*rm*(1.0 - s*s)*tp*vertices[57 + i] \
                  - 8.0*phys_x[i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void S2Jacobian3D(double* rcol,
                              double* scol,
                              double* tcol,
                              double* x,
                              double* vertices,
                              double* phys_x) nogil:
        cdef int i
        cdef double r, s, t, rm, rp, sm, sp, tm, tp

        r = x[0]
        rm = 1.0 - r
        rp = 1.0 + r

        s = x[1]
        sm = 1.0 - s
        sp = 1.0 + s

        t = x[2]
        tm = 1.0 - t
        tp = 1.0 + t

        for i in range(3):
            rcol[i] = (sm*tm*(r + s + t + 2.0) - rm*sm*tm)*vertices[0  + i] \
                    + (sm*tm*(r - s - t - 2.0) + rp*sm*tm)*vertices[3  + i] \
                    + (sp*tm*(r + s - t - 2.0) + rp*sp*tm)*vertices[6  + i] \
                    + (sp*tm*(r - s + t + 2.0) - rm*sp*tm)*vertices[9  + i] \
                    + (sm*tp*(r + s - t + 2.0) - rm*sm*tp)*vertices[12 + i] \
                    + (sm*tp*(r - s + t - 2.0) + rp*sm*tp)*vertices[15 + i] \
                    + (sp*tp*(r + s + t - 2.0) + rp*sp*tp)*vertices[18 + i] \
                    + (sp*tp*(r - s - t + 2.0) - rm*sp*tp)*vertices[21 + i] \
                    - 4.0*r*sm*tm*vertices[24 + i] \
                    + 2.0*(1.0 - s*s)*tm*vertices[27 + i] \
                    - 4.0*r*sp*tm*vertices[30 + i] \
                    - 2.0*(1.0 - s*s)*tm*vertices[33 + i] \
                    - 2.0*sm*(1.0 - t*t)*vertices[36 + i] \
                    + 2.0*sm*(1.0 - t*t)*vertices[39 + i] \
                    + 2.0*sp*(1.0 - t*t)*vertices[42 + i] \
                    - 2.0*sp*(1.0 - t*t)*vertices[45 + i] \
                    - 4.0*r*sm*tp*vertices[48 + i] \
                    + 2.0*(1.0 - s*s)*tp*vertices[51 + i] \
                    - 4.0*r*sp*tp*vertices[54 + i] \
                    - 2.0*(1.0 - s*s)*tp*vertices[57 + i]
            scol[i] = ( rm*tm*(r + s + t + 2.0) - rm*sm*tm)*vertices[0  + i] \
                    + (-rp*tm*(r - s - t - 2.0) - rp*sm*tm)*vertices[3  + i] \
                    + ( rp*tm*(r + s - t - 2.0) + rp*sp*tm)*vertices[6  + i] \
                    + (-rm*tm*(r - s + t + 2.0) + rm*sp*tm)*vertices[9  + i] \
                    + ( rm*tp*(r + s - t + 2.0) - rm*sm*tp)*vertices[12 + i] \
                    + (-rp*tp*(r - s + t - 2.0) - rp*sm*tp)*vertices[15 + i] \
                    + ( rp*tp*(r + s + t - 2.0) + rp*sp*tp)*vertices[18 + i] \
                    + (-rm*tp*(r - s - t + 2.0) + rm*sp*tp)*vertices[21 + i] \
                    - 2.0*(1.0 - r*r)*tm*vertices[24 + i] \
                    - 4.0*rp*s*tm*vertices[27 + i] \
                    + 2.0*(1.0 - r*r)*tm*vertices[30 + i] \
                    - 4.0*rm*s*tm*vertices[33 + i] \
                    - 2.0*rm*(1.0 - t*t)*vertices[36 + i] \
                    - 2.0*rp*(1.0 - t*t)*vertices[39 + i] \
                    + 2.0*rp*(1.0 - t*t)*vertices[42 + i] \
                    + 2.0*rm*(1.0 - t*t)*vertices[45 + i] \
                    - 2.0*(1.0 - r*r)*tp*vertices[48 + i] \
                    - 4.0*rp*s*tp*vertices[51 + i] \
                    + 2.0*(1.0 - r*r)*tp*vertices[54 + i] \
                    - 4.0*rm*s*tp*vertices[57 + i]
            tcol[i] = ( rm*sm*(r + s + t + 2.0) - rm*sm*tm)*vertices[0  + i] \
                    + (-rp*sm*(r - s - t - 2.0) - rp*sm*tm)*vertices[3  + i] \
                    + (-rp*sp*(r + s - t - 2.0) - rp*sp*tm)*vertices[6  + i] \
                    + ( rm*sp*(r - s + t + 2.0) - rm*sp*tm)*vertices[9  + i] \
                    + (-rm*sm*(r + s - t + 2.0) + rm*sm*tp)*vertices[12 + i] \
                    + ( rp*sm*(r - s + t - 2.0) + rp*sm*tp)*vertices[15 + i] \
                    + ( rp*sp*(r + s + t - 2.0) + rp*sp*tp)*vertices[18 + i] \
                    + (-rm*sp*(r - s - t + 2.0) + rm*sp*tp)*vertices[21 + i] \
                    - 2.0*(1.0 - r*r)*sm*vertices[24 + i] \
                    - 2.0*rp*(1.0 - s*s)*vertices[27 + i] \
                    - 2.0*(1.0 - r*r)*sp*vertices[30 + i] \
                    - 2.0*rm*(1.0 - s*s)*vertices[33 + i] \
                    - 4.0*rm*sm*t*vertices[36 + i] \
                    - 4.0*rp*sm*t*vertices[39 + i] \
                    - 4.0*rp*sp*t*vertices[42 + i] \
                    - 4.0*rm*sp*t*vertices[45 + i] \
                    + 2.0*(1.0 - r*r)*sm*vertices[48 + i] \
                    + 2.0*rp*(1.0 - s*s)*vertices[51 + i] \
                    + 2.0*(1.0 - r*r)*sp*vertices[54 + i] \
                    + 2.0*rm*(1.0 - s*s)*vertices[57 + i]


cdef class W1Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, linear, wedge mesh element.

    '''

    def __init__(self):
        super(W1Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = W1Function3D
        self.jac = W1Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F

        F = vals[0]*(1.0 - coord[0] - coord[1])*(1.0 - coord[2]) + \
            vals[1]*coord[0]*(1.0 - coord[2]) + \
            vals[2]*coord[1]*(1.0 - coord[2]) + \
            vals[3]*(1.0 - coord[0] - coord[1])*(1.0 + coord[2]) + \
            vals[4]*coord[0]*(1.0 + coord[2]) + \
            vals[5]*coord[1]*(1.0 + coord[2])

        return F / 2.0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for wedges the bounds of the mapped coordinates are:
        #     0 <= mapped_coord[0] <= 1 - mapped_coord[1]
        #     0 <= mapped_coord[1]
        #    -1 <= mapped_coord[2] <= 1
        if (mapped_coord[0] < -self.inclusion_tol or
            mapped_coord[0] + mapped_coord[1] - 1.0 > self.inclusion_tol):
            return 0
        if (mapped_coord[1] < -self.inclusion_tol):
            return 0
        if (fabs(mapped_coord[2]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        cdef double r, s, t
        cdef double thresh = 5.0e-2
        r = mapped_coord[0]
        s = mapped_coord[1]
        t = mapped_coord[2]

        cdef int near_edge_r, near_edge_s, near_edge_t
        near_edge_r = (r < thresh) or (fabs(r + s - 1.0) < thresh)
        near_edge_s = (s < thresh)
        near_edge_t = fabs(fabs(mapped_coord[2]) - 1.0) < thresh

        # we use ray.instID to pass back whether the ray is near the
        # element boundary or not (used to annotate mesh lines)
        if (near_edge_r and near_edge_s):
            return 1
        elif (near_edge_r and near_edge_t):
            return 1
        elif (near_edge_s and near_edge_t):
            return 1
        else:
            return -1


cdef class NonlinearSolveSampler2D(ElementSampler):

    '''

    This is a base class for handling element samplers that require
    a nonlinear solve to invert the mapping between coordinate systems.
    To do this, we perform Newton-Raphson iteration using a specified
    system of equations with an analytic Jacobian matrix. This solver
    is hard-coded for 2D for reasons of efficiency. This is not to be
    used directly, use one of the subclasses instead.

    '''

    def __init__(self):
        super(NonlinearSolveSampler2D, self).__init__()
        self.tolerance = 1.0e-9
        self.max_iter = 10

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self,
                               double* mapped_x,
                               double* vertices,
                               double* physical_x) nogil:
        '''

        A thorough description of Newton's method and modifications for global
        convergence can be found in Dennis's text "Numerical Methods for
        Unconstrained Optimization and Nonlinear Equations."

        x: solution vector; holds unit/mapped coordinates
        xk: temporary vector for holding solution of current iteration
        f: residual vector
        A: Jacobian matrix (derivative of residual vector wrt x)
        d: Jacobian determinant
        s_n: Newton step vector
        lam: fraction of Newton step by which to change x
        alpha: constant proportional to how much residual required to decrease.
               1e-4 is value of alpha recommended by Dennis
        err_c: Error of current iteration
        err_plus: Error of next iteration
        min_lam: minimum fraction of Newton step that the line search is allowed
                 to take. General experience suggests that lambda values smaller
                 than 1e-3 will not significantly reduce the residual, but we
                 set to 1e-6 just to be safe
        '''
        
        cdef int i
        cdef double d, lam
        cdef double[2] f
        cdef double[2] x, xk, s_n
        cdef double[4] A
        cdef int iterations = 0
        cdef double err_c, err_plus
        cdef double alpha = 1e-4
        cdef double min_lam = 1e-6

        # initial guess
        for i in range(2):
            x[i] = 0.0

        # initial error norm
        self.func(f, x, vertices, physical_x)
        err_c = maxnorm(f, 2)

        # begin Newton iteration
        while (err_c > self.tolerance and iterations < self.max_iter):
            self.jac(&A[0], &A[2], x, vertices, physical_x)
            d = (A[0]*A[3] - A[1]*A[2])

            s_n[0] = -( A[3]*f[0] - A[2]*f[1]) / d
            s_n[1] = -(-A[1]*f[0] + A[0]*f[1]) / d
            xk[0] = x[0] + s_n[0]
            xk[1] = x[1] + s_n[1]
            self.func(f, xk, vertices, physical_x)
            err_plus = maxnorm(f, 2)
            
            lam = 1
            while err_plus > err_c * (1. - alpha * lam) and lam > min_lam:
                lam = lam / 2
                xk[0] = x[0] + lam * s_n[0]
                xk[1] = x[1] + lam * s_n[1]
                self.func(f, xk, vertices, physical_x)
                err_plus = maxnorm(f, 2)

            x[0] = xk[0]
            x[1] = xk[1]
            err_c = err_plus
            iterations += 1

        if (err_c > self.tolerance):
            # we did not converge, set bogus value
            for i in range(2):
                mapped_x[i] = -99.0
        else:
            for i in range(2):
                mapped_x[i] = x[i]


cdef class Q1Sampler2D(NonlinearSolveSampler2D):

    '''

    This implements sampling inside a 2D, linear, quadrilateral mesh element.

    '''

    def __init__(self):
        super(Q1Sampler2D, self).__init__()
        self.num_mapped_coords = 2
        self.dim = 2
        self.func = Q1Function2D
        self.jac = Q1Jacobian2D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, rm, rp, sm, sp

        rm = 1.0 - coord[0]
        rp = 1.0 + coord[0]
        sm = 1.0 - coord[1]
        sp = 1.0 + coord[1]

        F = vals[0]*rm*sm + vals[1]*rp*sm + vals[2]*rp*sp + vals[3]*rm*sp
        return 0.25*F

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for quads, we check whether the mapped_coord is between
        # -1 and 1 in both directions.
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

cdef class Q2Sampler2D(NonlinearSolveSampler2D):

    '''

    This implements sampling inside a 2D, quadratic, quadrilateral mesh element.

    '''

    def __init__(self):
        super(Q2Sampler2D, self).__init__()
        self.num_mapped_coords = 2
        self.dim = 2
        self.func = Q2Function2D
        self.jac = Q2Jacobian2D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, rm, rp, sm, sp
        cdef double[9] phi
        cdef double rv = 0

        zet = coord[0]
        eta = coord[1]
        zetm = coord[0] - 1.
        zetp = coord[0] + 1.
        etam = coord[1] - 1.
        etap = coord[1] + 1.

        phi[0] = zet * zetm * eta * etam / 4.
        phi[1] = zet * zetp * eta * etam / 4.
        phi[2] = zet * zetp * eta * etap / 4.
        phi[3] = zet * zetm * eta * etap / 4.
        phi[4] = zetp * zetm * eta * etam / -2.
        phi[5] = zet * zetp * etap * etam / -2.
        phi[6] = zetp * zetm * eta * etap / -2.
        phi[7] = zet * zetm * etap * etam / -2.
        phi[8] = zetp * zetm * etap * etam

        for i in range(9):
            rv += vals[i] * phi[i]

        return rv

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for quads, we check whether the mapped_coord is between
        # -1 and 1 in both directions.
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    
cdef class T2Sampler2D(NonlinearSolveSampler2D):

    '''

    This implements sampling inside a 2D, quadratic, triangular mesh
    element. Note that this implementation uses canonical coordinates.

    '''

    def __init__(self):
        super(T2Sampler2D, self).__init__()
        self.num_mapped_coords = 2
        self.dim = 2
        self.func = T2Function2D
        self.jac = T2Jacobian2D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double phi0, phi1, phi2, phi3, phi4, phi5, c0sq, c1sq, c0c1

        c0sq = coord[0] * coord[0]
        c1sq = coord[1] * coord[1]
        c0c1 = coord[0] * coord[1]

        phi0 = 1 - 3 * coord[0] + 2 * c0sq - 3 * coord[1] + \
               2 * c1sq + 4 * c0c1
        phi1 = -coord[0] + 2 * c0sq
        phi2 = -coord[1] + 2 * c1sq
        phi3 = 4 * coord[0] - 4 * c0sq - 4 * c0c1
        phi4 = 4 * c0c1
        phi5 = 4 * coord[1] - 4 * c1sq - 4 * c0c1

        return vals[0]*phi0 + vals[1]*phi1 + vals[2]*phi2 + vals[3]*phi3 + \
               vals[4]*phi4 + vals[5]*phi5

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for canonical tris, we check whether the mapped_coords are between
        # 0 and 1.
        if (mapped_coord[0] < -self.inclusion_tol or \
            mapped_coord[1] < -self.inclusion_tol or \
            mapped_coord[0] + mapped_coord[1] - 1.0 > self.inclusion_tol):
            return 0
        return 1

cdef class Tet2Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, quadratic, tetrahedral mesh
    element. Note that this implementation uses canonical coordinates.

    '''

    def __init__(self):
        super(Tet2Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = Tet2Function3D
        self.jac = Tet2Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double[10] phi
        cdef double coordsq[3]
        cdef int i
        cdef double return_value = 0

        for i in range(3):
            coordsq[i] = coord[i] * coord[i]

        phi[0] = 1 - 3 * coord[0] + 2 * coordsq[0] - 3 * coord[1] + \
                 2 * coordsq[1] - 3 * coord[2] + 2 * coordsq[2] + \
                 4 * coord[0] * coord[1] + 4 * coord[0] * coord[2] + \
                 4 * coord[1] * coord[2]
        phi[1] = -coord[0] + 2 * coordsq[0]
        phi[2] = -coord[1] + 2 * coordsq[1]
        phi[3] = -coord[2] + 2 * coordsq[2]
        phi[4] = 4 * coord[0] - 4 * coordsq[0] - 4 * coord[0] * coord[1] - \
                 4 * coord[0] * coord[2]
        phi[5] = 4 * coord[0] * coord[1]
        phi[6] = 4 * coord[1] - 4 * coordsq[1] - 4 * coord[0] * coord[1] - \
                 4 * coord[1] * coord[2]
        phi[7] = 4 * coord[2] - 4 * coordsq[2] - 4 * coord[2] * coord[0] - \
                 4 * coord[2] * coord[1]
        phi[8] = 4 * coord[0] * coord[2]
        phi[9] = 4 * coord[1] * coord[2]

        for i in range(10):
            return_value += phi[i] * vals[i]

        return return_value

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for canonical tets, we check whether the mapped_coords are between
        # 0 and 1.
        if (mapped_coord[0] < -self.inclusion_tol or \
            mapped_coord[1] < -self.inclusion_tol or \
            mapped_coord[2] < -self.inclusion_tol or \
            mapped_coord[0] + mapped_coord[1] + mapped_coord[2] - 1.0 > \
            self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        cdef double u, v
        cdef double thresh = 2.0e-2
        if mapped_coord[0] == 0:
            u = mapped_coord[1]
            v = mapped_coord[2]
        elif mapped_coord[1] == 0:
            u = mapped_coord[2]
            v = mapped_coord[0]
        elif mapped_coord[2] == 0:
            u = mapped_coord[1]
            v = mapped_coord[0]
        else:
            u = mapped_coord[1]
            v = mapped_coord[2]
        if ((u < thresh) or
            (v < thresh) or
            (fabs(u - 1) < thresh) or
            (fabs(v - 1) < thresh)):
            return 1
        return -1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_hex_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                     np.ndarray[np.float64_t, ndim=1] field_values,
                     np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    cdef Q1Sampler3D sampler = Q1Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_hex20_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    cdef S2Sampler3D sampler = S2Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tetra_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = P1Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_wedge_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    cdef W1Sampler3D sampler = W1Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_linear1D_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                          np.ndarray[np.float64_t, ndim=1] field_values,
                          np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = P1Sampler1D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tri_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                     np.ndarray[np.float64_t, ndim=1] field_values,
                     np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = P1Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_quad_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                      np.ndarray[np.float64_t, ndim=1] field_values,
                      np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = Q1Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_quad2_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = Q2Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tri2_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                      np.ndarray[np.float64_t, ndim=1] field_values,
                      np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = T2Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tet2_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                      np.ndarray[np.float64_t, ndim=1] field_values,
                      np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = Tet2Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val



from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

class libconfig(dict):
    def __init__(self, config=None):
        if config is not None:
            self.read(config)

    def read(self, config):
        if not hasattr(config, "read"):
            cfile = open(config, 'r')
        else:
            cfile = config

        # Strip out spaces and blanks
        lines = [line.strip() for line in cfile if len(line) > 0]

        # Strip out comments
        lines = [line for line in lines if not (line.startswith('#') or
                                                line.startswith('/'))]

        # Concatenate
        oneline = ''
        for line in lines:
            oneline += line

        statements = oneline.split(';')

        self.parse_statements(self, statements)

    def parse_statements(self, this_dict, statements):
        while len(statements) > 0:
            statement = statements.pop(0)
            if len(statement) == 0:
                continue
            if ':' in statement:
                # DICTIONARY
                new_level_lines = []
                k, v = statement.split(':', 1)
                k = k.strip(' :')
                v = v.strip(' {')
                level = 1 + v.count(':')
                this_dict[k] = {}
                new_level_lines.append(v)

                while(level > 0 and len(statements) > 0):
                    nextline = statements.pop(0).strip()
                    level += nextline.count('{')
                    if nextline == '}' and level == 1:
                        level = 0
                        break
                    new_level_lines.append(nextline)
                    level -= nextline.count('}')
                self.parse_statements(this_dict[k], new_level_lines)
            else:
                k, v = statement.split('=')
                k = k.strip()
                v = v.strip()
                this_dict[k] = self.correct_type(v)

    def correct_type(self, v):
        if v == "true":
            v = "True"
        elif v == "false":
            v = "False"
        # ...are we really evaling this?  We should work around that somehow.
        return eval(v)

    def write(self, filename):
        f = open(filename, 'w')

        self.write_dict(f, self, 0)

    def write_dict(self, f, this_dict, level):
        tab = ' '*4

        dict_dict = {}
        for k, v in this_dict.items():
            if type(v) == dict:
                dict_dict[k] = v
            else:
                if type(v) == str:
                    v = '"%s"' % v
                f.writelines(tab*level + '%s = %s;\n' % (k, v))

        for k, v in dict_dict.items():
            f.writelines('\n')
            f.writelines(tab*level + '%s :\n' % k)
            f.writelines(tab*level+'{\n')
            self.write_dict(f, v, level+1)
            f.writelines(tab*level+'};\n')

if __name__ == '__main__':
    cfg = libconfig()
    cfg.read('test_config.cfg')
    print(cfg)



"""
API for yt.frontends.fits
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
      FITSGrid, \
      FITSHierarchy, \
      FITSDataset

from .fields import \
      FITSFieldInfo

from .io import \
      IOHandlerFITS

from .misc import \
      setup_counts_fields

from . import tests



"""
Create randomly centered, tiled projections to be used in light cones.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.funcs import \
     mylog
from yt.units.yt_array import \
    uconcatenate
from yt.visualization.fixed_resolution import \
    FixedResolutionBuffer
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_blocking_call

@parallel_blocking_call
def _light_cone_projection(my_slice, field, pixels, weight_field=None,
                           save_image=False, field_cuts=None):
    "Create a single projection to be added into the light cone stack."

    # We are just saving the projection object, so only the projection axis 
    # needs to be considered since the lateral shifting and tiling occurs after 
    # the projection object is made.
    # Likewise, only the box_depth_fraction needs to be considered.

    mylog.info("Making projection at z = %f from %s." % \
               (my_slice["redshift"], my_slice["filename"]))

    region_center = [0.5 * (my_slice["object"].domain_right_edge[q] +
                            my_slice["object"].domain_left_edge[q]) \
                         for q in range(my_slice["object"].dimensionality)]

    # 1. The Depth Problem
    # Use coordinate field cut in line of sight to cut projection to proper depth.
    if field_cuts is None:
        these_field_cuts = []
    else:
        these_field_cuts = field_cuts.copy()

    if (my_slice["box_depth_fraction"] < 1):
        axis = ("x", "y", "z")[my_slice["projection_axis"]]
        depthLeft = \
          my_slice["projection_center"][my_slice["projection_axis"]] \
            - 0.5 * my_slice["box_depth_fraction"]
        depthRight = \
          my_slice["projection_center"][my_slice["projection_axis"]] \
            + 0.5 * my_slice["box_depth_fraction"]
        if (depthLeft < 0):
            cut_mask = ("((obj[\"%s\"] + 0.5*obj[\"d%s\"] >= 0) & " + \
              "(obj[\"%s\"] - 0.5*obj[\"d%s\"] <= %f)) | " + \
              "((obj[\"%s\"] + 0.5*obj[\"d%s\"] >= %f) & " + \
              "(obj[\"%s\"] - 0.5*obj[\"d%s\"] <= 1))") % \
                (axis, axis, axis, axis, depthRight, 
                 axis, axis, (depthLeft+1), axis, axis)
        elif (depthRight > 1):
            cut_mask = ("((obj[\"%s\"] + 0.5*obj[\"d%s\"] >= 0) & " + \
              "(obj[\"%s\"] - 0.5*obj[\"d%s\"] <= %f)) | " + \
              "((obj[\"%s\"] + 0.5*obj[\"d%s\"] >= %f) & " + \
              "(obj[\"%s\"] - 0.5*obj[\"d%s\"] <= 1))") % \
                (axis, axis, axis, axis, (depthRight-1),
                 axis, axis, depthLeft, axis, axis)
        else:
            cut_mask = ("(obj[\"%s\"] + 0.5*obj[\"d%s\"] >= %f) & " + \
              "(obj[\"%s\"] - 0.5*obj[\"%s\"] <= %f)") % \
              (axis, axis, depthLeft, axis, axis, depthRight)

        these_field_cuts.append(cut_mask)

    data_source = my_slice["object"].all_data()
    cut_region = data_source.cut_region(these_field_cuts)
        
    # Make projection.
    proj = my_slice["object"].proj(field, my_slice["projection_axis"], 
        weight_field, center=region_center,
        data_source=cut_region)
    proj_field = proj.field[0]

    del data_source, cut_region
    
    # 2. The Tile Problem
    # Tile projection to specified width.

    # Original projection data.
    original_px = proj.field_data["px"].in_units("code_length").copy()
    original_py = proj.field_data["py"].in_units("code_length").copy()
    original_pdx = proj.field_data["pdx"].in_units("code_length").copy()
    original_pdy = proj.field_data["pdy"].in_units("code_length").copy()
    original_field = proj.field_data[proj_field].copy()
    original_weight_field = proj.field_data["weight_field"].copy()

    for my_field in ["px", "py", "pdx", "pdy", proj_field, "weight_field"]:
        proj.field_data[my_field] = [proj.field_data[my_field]]

    # Copy original into offset positions to make tiles.
    for x in range(int(np.ceil(my_slice["box_width_fraction"]))):
        x = my_slice["object"].quan(x, "code_length")
        for y in range(int(np.ceil(my_slice["box_width_fraction"]))):
            y = my_slice["object"].quan(y, "code_length")
            if ((x + y) > 0):
                proj.field_data["px"] += [original_px+x]
                proj.field_data["py"] += [original_py+y]
                proj.field_data["pdx"] += [original_pdx]
                proj.field_data["pdy"] += [original_pdy]
                proj.field_data["weight_field"] += [original_weight_field]
                proj.field_data[proj_field] += [original_field]

    for my_field in ["px", "py", "pdx", "pdy", proj_field, "weight_field"]:
        proj.field_data[my_field] = \
          my_slice["object"].arr(proj.field_data[my_field]).flatten()

    # Delete originals.
    del original_px
    del original_py
    del original_pdx
    del original_pdy
    del original_field
    del original_weight_field

    # 3. The Shift Problem
    # Shift projection by random x and y offsets.

    image_axes = np.roll(np.arange(3), -my_slice["projection_axis"])[1:]
    di_left_x  = my_slice["object"].domain_left_edge[image_axes[0]]
    di_right_x = my_slice["object"].domain_right_edge[image_axes[0]]
    di_left_y  = my_slice["object"].domain_left_edge[image_axes[1]]
    di_right_y = my_slice["object"].domain_right_edge[image_axes[1]]
    
    offset = my_slice["projection_center"].copy() * \
      my_slice["object"].domain_width
    offset = np.roll(offset, -my_slice["projection_axis"])[1:]

    # Shift x and y positions.
    proj.field_data["px"] -= offset[0]
    proj.field_data["py"] -= offset[1]

    # Wrap off-edge cells back around to other side (periodic boundary conditions).
    proj.field_data["px"][proj.field_data["px"] < di_left_x] += \
      np.ceil(my_slice["box_width_fraction"]) * di_right_x
    proj.field_data["py"][proj.field_data["py"] < di_left_y] += \
      np.ceil(my_slice["box_width_fraction"]) * di_right_y

    # After shifting, some cells have fractional coverage on both sides of the box.
    # Find those cells and make copies to be placed on the other side.

    # Cells hanging off the right edge.
    add_x_right = proj.field_data["px"] + 0.5 * proj.field_data["pdx"] > \
      np.ceil(my_slice["box_width_fraction"]) * di_right_x
    add_x_px = proj.field_data["px"][add_x_right]
    add_x_px -= np.ceil(my_slice["box_width_fraction"]) * di_right_x
    add_x_py = proj.field_data["py"][add_x_right]
    add_x_pdx = proj.field_data["pdx"][add_x_right]
    add_x_pdy = proj.field_data["pdy"][add_x_right]
    add_x_field = proj.field_data[proj_field][add_x_right]
    add_x_weight_field = proj.field_data["weight_field"][add_x_right]
    del add_x_right

    # Cells hanging off the left edge.
    add_x_left = proj.field_data["px"] - 0.5 * proj.field_data["pdx"] < di_left_x
    add2_x_px = proj.field_data["px"][add_x_left]
    add2_x_px += np.ceil(my_slice["box_width_fraction"]) * di_right_x
    add2_x_py = proj.field_data["py"][add_x_left]
    add2_x_pdx = proj.field_data["pdx"][add_x_left]
    add2_x_pdy = proj.field_data["pdy"][add_x_left]
    add2_x_field = proj.field_data[proj_field][add_x_left]
    add2_x_weight_field = proj.field_data["weight_field"][add_x_left]
    del add_x_left

    # Cells hanging off the top edge.
    add_y_right = proj.field_data["py"] + 0.5 * proj.field_data["pdy"] > \
      np.ceil(my_slice["box_width_fraction"]) * di_right_y
    add_y_px = proj.field_data["px"][add_y_right]
    add_y_py = proj.field_data["py"][add_y_right]
    add_y_py -= np.ceil(my_slice["box_width_fraction"]) * di_right_y
    add_y_pdx = proj.field_data["pdx"][add_y_right]
    add_y_pdy = proj.field_data["pdy"][add_y_right]
    add_y_field = proj.field_data[proj_field][add_y_right]
    add_y_weight_field = proj.field_data["weight_field"][add_y_right]
    del add_y_right

    # Cells hanging off the bottom edge.
    add_y_left = proj.field_data["py"] - 0.5 * proj.field_data["pdy"] < di_left_y
    add2_y_px = proj.field_data["px"][add_y_left]
    add2_y_py = proj.field_data["py"][add_y_left]
    add2_y_py += np.ceil(my_slice["box_width_fraction"]) * di_right_y
    add2_y_pdx = proj.field_data["pdx"][add_y_left]
    add2_y_pdy = proj.field_data["pdy"][add_y_left]
    add2_y_field = proj.field_data[proj_field][add_y_left]
    add2_y_weight_field = proj.field_data["weight_field"][add_y_left]
    del add_y_left

    # Add the hanging cells back to the projection data.
    proj.field_data["px"] = uconcatenate(
        [proj.field_data["px"], add_x_px,
         add_y_px, add2_x_px, add2_y_px])
    proj.field_data["py"] = uconcatenate(
        [proj.field_data["py"], add_x_py,
         add_y_py, add2_x_py, add2_y_py])
    proj.field_data["pdx"] = uconcatenate(
        [proj.field_data["pdx"], add_x_pdx,
         add_y_pdx, add2_x_pdx, add2_y_pdx])
    proj.field_data["pdy"] = uconcatenate(
        [proj.field_data["pdy"], add_x_pdy,
         add_y_pdy, add2_x_pdy, add2_y_pdy])
    proj.field_data[proj_field] = uconcatenate(
        [proj.field_data[proj_field], add_x_field,
         add_y_field, add2_x_field, add2_y_field])
    proj.field_data["weight_field"] = uconcatenate(
        [proj.field_data["weight_field"],
         add_x_weight_field, add_y_weight_field,
         add2_x_weight_field, add2_y_weight_field])

    # Delete original copies of hanging cells.
    del add_x_px, add_y_px, add2_x_px, add2_y_px
    del add_x_py, add_y_py, add2_x_py, add2_y_py
    del add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx
    del add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy
    del add_x_field, add_y_field, add2_x_field, add2_y_field
    del add_x_weight_field, add_y_weight_field, add2_x_weight_field, add2_y_weight_field

    # Tiles were made rounding up the width to the nearest integer.
    # Cut off the edges to get the specified width.
    # Cut in the x direction.
    cut_x = proj.field_data["px"] - 0.5 * proj.field_data["pdx"] < \
      di_right_x * my_slice["box_width_fraction"]
    proj.field_data["px"] = proj.field_data["px"][cut_x]
    proj.field_data["py"] = proj.field_data["py"][cut_x]
    proj.field_data["pdx"] = proj.field_data["pdx"][cut_x]
    proj.field_data["pdy"] = proj.field_data["pdy"][cut_x]
    proj.field_data[proj_field] = proj.field_data[proj_field][cut_x]
    proj.field_data["weight_field"] = proj.field_data["weight_field"][cut_x]
    del cut_x

    # Cut in the y direction.
    cut_y = proj.field_data["py"] - 0.5 * proj.field_data["pdy"] < \
      di_right_y * my_slice["box_width_fraction"]
    proj.field_data["px"] = proj.field_data["px"][cut_y]
    proj.field_data["py"] = proj.field_data["py"][cut_y]
    proj.field_data["pdx"] = proj.field_data["pdx"][cut_y]
    proj.field_data["pdy"] = proj.field_data["pdy"][cut_y]
    proj.field_data[proj_field] = proj.field_data[proj_field][cut_y]
    proj.field_data["weight_field"] = proj.field_data["weight_field"][cut_y]
    del cut_y

    # Create fixed resolution buffer to return back to the light cone object.
    # These buffers will be stacked together to make the light cone.
    frb = FixedResolutionBuffer(proj, 
        (di_left_x, di_right_x * my_slice["box_width_fraction"],
         di_left_y, di_right_y * my_slice["box_width_fraction"]),
        (pixels, pixels), antialias=False)

    return frb



import numpy as np

def enhance(im, stdval=6.0, just_alpha=True):
    if just_alpha:
        nz = im[im>0.0]
        im[:] = im[:]/(nz.mean()+stdval*np.std(nz))
    else:
        for c in range(3):
            nz = im[:,:,c][im[:,:,c]>0.0]
            im[:,:,c] = im[:,:,c]/(nz.mean()+stdval*np.std(nz))
            del nz
    np.clip(im, 0.0, 1.0, im)

def enhance_rgba(im, stdval=6.0):
    nzc = im[:,:,:3][im[:,:,:3]>0.0]
    cmax = nzc.mean()+stdval*nzc.std()


    nza = im[:,:,3][im[:,:,3]>0.0]
    if len(nza) == 0:
        im[:,:,3]=1.0
        amax = 1.0
    else:
        amax = nza.mean()+stdval*nza.std()

    im.rescale(amax=amax, cmax=cmax, inline=True)
    np.clip(im, 0.0, 1.0, im)




"""
Miscellaneous functions that are Enzo-P-specific



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

def bdecode(block):
    if ":" in block:
        level = len(block) - block.find(":") - 1
    else:
        level = 0
    bst   = block.replace(":", "")
    d     = float(2**len(bst))
    left  = int(bst, 2)
    right = left + 1
    left  /= d
    right /= d
    return level, left, right

def get_block_info(block, min_dim=3):
    mybs = block[1:].split("_")
    dim = max(len(mybs), min_dim)
    left = np.zeros(dim)
    right = np.ones(dim)
    for i, myb in enumerate(mybs):
        level, left[i], right[i] = bdecode(myb)
    return level, left, right

def get_root_blocks(block, min_dim=3):
    mybs = block[1:].split("_")
    dim = max(len(mybs), min_dim)
    nb = np.ones(dim, dtype=int)
    for i, myb in enumerate(mybs):
        if ":" in myb:
            s = myb.find(":")
        else:
            s = len(myb)
        nb[i] = 2**s
    return nb

def get_root_block_id(block, min_dim=3):
    mybs = block[1:].split("_")
    dim = max(len(mybs), min_dim)
    rbid = np.zeros(dim, dtype=int)
    for i, myb in enumerate(mybs):
        if ":" in myb:
            s = myb.find(":")
        else:
            s = len(myb)
        rbid[i] = int(myb[:s], 2)
    return rbid

def get_child_index(anc_id, desc_id):
    cid = ""
    for aind, dind in zip( anc_id.split("_"),
                          desc_id.split("_")):
        cid += dind[len(aind)]
    cid = int(cid, 2)
    return cid

def is_parent(anc_block, desc_block):
    for aind, dind in zip( anc_block.split("_"),
                          desc_block.split("_")):
        if not dind.startswith(aind):
            return False
    return True



import glob
import os

from yt.config import ytcfg
from yt.data_objects.time_series import DatasetSeries
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    GenericArrayTest

def setup():
    ytcfg["yt","__withintesting"] = "True"

data_path = ytcfg.get("yt", "test_data_dir")

pfields = ["particle_position_x", "particle_position_y", "particle_position_z"]
vfields = ["particle_velocity_x", "particle_velocity_y", "particle_velocity_z"]

@requires_ds("Orbit/orbit_hdf5_chk_0000")
def test_orbit_traj():
    fields = ["particle_velocity_x", "particle_velocity_y", "particle_velocity_z"]
    my_fns = glob.glob(os.path.join(data_path, "Orbit/orbit_hdf5_chk_00[0-9][0-9]"))
    my_fns.sort()
    ts = DatasetSeries(my_fns)
    ds = ts[0]
    traj = ts.particle_trajectories([1, 2], fields=fields, suppress_logging=True)
    for field in pfields+vfields:
        def field_func(name):
            return traj[field]
        yield GenericArrayTest(ds, field_func, args=[field])

@requires_ds("enzo_tiny_cosmology/DD0000/DD0000")
def test_etc_traj():
    fields = ["particle_velocity_x", "particle_velocity_y", "particle_velocity_z"]
    my_fns = glob.glob(os.path.join(data_path, "enzo_tiny_cosmology/DD000[0-9]/*.hierarchy"))
    my_fns.sort()
    ts = DatasetSeries(my_fns)
    ds = ts[0]
    sp = ds.sphere("max", (0.5, "Mpc"))
    indices = sp["particle_index"][sp["particle_type"] == 1][:5]
    traj = ts.particle_trajectories(indices, fields=fields, suppress_logging=True)
    traj.add_fields(["density"])
    for field in pfields+vfields+["density"]:
        def field_func(name):
            return traj[field]
        yield GenericArrayTest(ds, field_func, args=[field])



"""
ImageArray Class

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import warnings
import numpy as np
from yt.config import \
    ytcfg
from yt.visualization.image_writer import write_bitmap, write_image
from yt.units.yt_array import YTArray


class ImageArray(YTArray):
    r"""A custom Numpy ndarray used for images.

    This differs from ndarray in that you can optionally specify an
    info dictionary which is used later in saving, and can be accessed with
    ImageArray.info.

    Parameters
    ----------
    input_array: array_like
        A numpy ndarray, or list.

    Other Parameters
    ----------------
    info: dictionary
        Contains information to be stored with image.

    Returns
    -------
    obj: ImageArray object

    Raises
    ------
    None

    See Also
    --------
    numpy.ndarray : Inherits

    Notes
    -----

    References
    ----------

    Examples
    --------
    These are written in doctest format, and should illustrate how to
    use the function.  Use the variables 'ds' for the dataset, 'pc' for
    a plot collection, 'c' for a center, and 'L' for a vector.

    >>> im = np.zeros([64,128,3])
    >>> for i in range(im.shape[0]):
    ...     for k in range(im.shape[2]):
    ...         im[i,:,k] = np.linspace(0.,0.3*k, im.shape[1])

    >>> myinfo = {'field':'dinosaurs', 'east_vector':np.array([1.,0.,0.]),
    ...     'north_vector':np.array([0.,0.,1.]), 'normal_vector':np.array([0.,1.,0.]),
    ...     'width':0.245, 'units':'cm', 'type':'rendering'}

    >>> im_arr = ImageArray(im, info=myinfo)
    >>> im_arr.save('test_ImageArray')

    Numpy ndarray documentation appended:

    """
    def __new__(cls, input_array, input_units=None, registry=None, info=None):
        obj = super(ImageArray, cls).__new__(cls, input_array, input_units, registry)
        if info is None:
            info = {}
        obj.info = info
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        super(ImageArray, self).__array_finalize__(obj)
        self.info = getattr(obj, 'info', None)

    def write_hdf5(self, filename, dataset_name=None):
        r"""Writes ImageArray to hdf5 file.

        Parameters
        ----------
        filename: string
        The filename to create and write a dataset to

        dataset_name: string
            The name of the dataset to create in the file.

        Examples
        --------
        >>> im = np.zeros([64,128,3])
        >>> for i in range(im.shape[0]):
        ...     for k in range(im.shape[2]):
        ...         im[i,:,k] = np.linspace(0.,0.3*k, im.shape[1])

        >>> myinfo = {'field':'dinosaurs', 'east_vector':np.array([1.,0.,0.]),
        ...     'north_vector':np.array([0.,0.,1.]), 'normal_vector':np.array([0.,1.,0.]),
        ...     'width':0.245, 'units':'cm', 'type':'rendering'}

        >>> im_arr = ImageArray(im, info=myinfo)
        >>> im_arr.write_hdf5('test_ImageArray.h5')

        """
        if dataset_name is None:
            dataset_name = self.info.get("name", "image")
        super(ImageArray, self).write_hdf5(filename, dataset_name=dataset_name,
                                           info=self.info)

    def add_background_color(self, background='black', inline=True):
        r"""Adds a background color to a 4-channel ImageArray

        This adds a background color to a 4-channel ImageArray, by default
        doing so inline.  The ImageArray must already be normalized to the
        [0,1] range.

        Parameters
        ----------
        background:
            This can be used to set a background color for the image, and can
            take several types of values:

               * ``white``: white background, opaque
               * ``black``: black background, opaque
               * ``None``: transparent background
               * 4-element array [r,g,b,a]: arbitrary rgba setting.

            Default: 'black'
        inline: boolean, optional
            If True, original ImageArray is modified. If False, a copy is first
            created, then modified. Default: True

        Returns
        -------
        out: ImageArray
            The modified ImageArray with a background color added.

        Examples
        --------
        >>> im = np.zeros([64,128,4])
        >>> for i in range(im.shape[0]):
        ...     for k in range(im.shape[2]):
        ...         im[i,:,k] = np.linspace(0.,10.*k, im.shape[1])

        >>> im_arr = ImageArray(im)
        >>> im_arr.rescale()
        >>> new_im = im_arr.add_background_color([1.,0.,0.,1.], inline=False)
        >>> new_im.write_png('red_bg.png')
        >>> im_arr.add_background_color('black')
        >>> im_arr.write_png('black_bg.png')
        """
        assert(self.shape[-1] == 4)

        if background is None:
            background = (0., 0., 0., 0.)
        elif background == 'white':
            background = (1., 1., 1., 1.)
        elif background == 'black':
            background = (0., 0., 0., 1.)

        # Alpha blending to background
        if inline:
            out = self
        else:
            out = self.copy()

        for i in range(3):
            out[:, :, i] = self[:, :, i]*self[:, :, 3]
            out[:, :, i] += background[i]*background[3]*(1.0-self[:, :, 3])
        out[:, :, 3] = self[:, :, 3]+background[3]*(1.0-self[:, :, 3])
        return out

    def rescale(self, cmax=None, amax=None, inline=True):
        r"""Rescales the image to be in [0,1] range.

        Parameters
        ----------
        cmax: float, optional
            Normalization value to use for rgb channels. Defaults to None,
            corresponding to using the maximum value in the rgb channels.
        amax: float, optional
            Normalization value to use for alpha channel. Defaults to None,
            corresponding to using the maximum value in the alpha channel.
        inline: boolean, optional
            Specifies whether or not the rescaling is done inline. If false,
            a new copy of the ImageArray will be created, returned.
            Default:True.

        Returns
        -------
        out: ImageArray
            The rescaled ImageArray, clipped to the [0,1] range.

        Notes
        -----
        This requires that the shape of the ImageArray to have a length of 3,
        and for the third dimension to be >= 3.  If the third dimension has
        a shape of 4, the alpha channel will also be rescaled.

        Examples
        --------
        >>> im = np.zeros([64,128,4])
        >>> for i in range(im.shape[0]):
        ...     for k in range(im.shape[2]):
        ...         im[i,:,k] = np.linspace(0.,0.3*k, im.shape[1])

        >>> im = ImageArray(im)
        >>> im.write_png('original.png')
        >>> im.rescale()
        >>> im.write_png('normalized.png')

        """
        assert(len(self.shape) == 3)
        assert(self.shape[2] >= 3)
        if inline:
            out = self
        else:
            out = self.copy()
        if cmax is None:
            cmax = self[:, :, :3].sum(axis=2).max()

        np.multiply(self[:, :, :3], 1.0/cmax, out[:, :, :3])

        if self.shape[2] == 4:
            if amax is None:
                amax = self[:, :, 3].max()
            if amax > 0.0:
                np.multiply(self[:, :, 3], 1.0/amax, out[:, :, 3])

        np.clip(out, 0.0, 1.0, out)
        return out

    def write_png(self, filename, sigma_clip=None, background='black',
                  rescale=True, clip_ratio=None):
        r"""Writes ImageArray to png file.

        Parameters
        ----------
        filename: string
            Filename to save to.  If None, PNG contents will be returned as a
            string.
        sigma_clip: float, optional
            Image will be clipped before saving to the standard deviation
            of the image multiplied by this value.  Useful for enhancing
            images. Default: None
        background:
            This can be used to set a background color for the image, and can
            take several types of values:

               * ``white``: white background, opaque
               * ``black``: black background, opaque
               * ``None``: transparent background
               * 4-element array [r,g,b,a]: arbitrary rgba setting.

            Default: 'black'
        rescale: boolean, optional
            If True, will write out a rescaled image (without modifying the
            original image). Default: True

        Examples
        --------
        >>> im = np.zeros([64,128,4])
        >>> for i in range(im.shape[0]):
        ...     for k in range(im.shape[2]):
        ...         im[i,:,k] = np.linspace(0.,10.*k, im.shape[1])

        >>> im_arr = ImageArray(im)
        >>> im_arr.write_png('standard.png')
        >>> im_arr.write_png('non-scaled.png', rescale=False)
        >>> im_arr.write_png('black_bg.png', background='black')
        >>> im_arr.write_png('white_bg.png', background='white')
        >>> im_arr.write_png('green_bg.png', background=[0,1,0,1])
        >>> im_arr.write_png('transparent_bg.png', background=None)

        """
        if rescale:
            scaled = self.rescale(inline=False)
        else:
            scaled = self

        if self.shape[-1] == 4:
            out = scaled.add_background_color(background, inline=False)
        else:
            out = scaled

        if filename is not None and filename[-4:] != '.png':
            filename += '.png'

        if clip_ratio is not None:
            warnings.warn("'clip_ratio' keyword is deprecated. Use 'sigma_clip' instead")
            sigma_clip = clip_ratio

        if sigma_clip is not None:
            nz = out[:, :, :3][out[:, :, :3].nonzero()]
            return write_bitmap(out.swapaxes(0, 1), filename,
                                nz.mean() + sigma_clip * nz.std())
        else:
            return write_bitmap(out.swapaxes(0, 1), filename)

    def write_image(self, filename, color_bounds=None, channel=None,
                    cmap_name=None, func=lambda x: x):
        r"""Writes a single channel of the ImageArray to a png file.

        Parameters
        ----------
        filename: string
            Note filename not be modified.

        Other Parameters
        ----------------
        channel: int
            Which channel to write out as an image. Defaults to 0
        cmap_name: string
            Name of the colormap to be used.
        color_bounds : tuple of floats, optional
            The min and max to scale between.  Outlying values will be clipped.
        cmap_name : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps .
        func : function, optional
            A function to transform the buffer before applying a colormap.

        Returns
        -------
        scaled_image : uint8 image that has been saved

        Examples
        --------

        >>> im = np.zeros([64,128])
        >>> for i in range(im.shape[0]):
        ...     im[i,:] = np.linspace(0.,0.3*i, im.shape[1])

        >>> myinfo = {'field':'dinosaurs', 'east_vector':np.array([1.,0.,0.]),
        ...     'north_vector':np.array([0.,0.,1.]), 'normal_vector':np.array([0.,1.,0.]),
        ...     'width':0.245, 'units':'cm', 'type':'rendering'}

        >>> im_arr = ImageArray(im, info=myinfo)
        >>> im_arr.write_image('test_ImageArray.png')

        """
        if cmap_name is None:
            cmap_name = ytcfg.get("yt", "default_colormap")
        if filename[-4:] != '.png':
            filename += '.png'

        #TODO: Write info dict as png metadata
        if channel is None:
            return write_image(self.swapaxes(0, 1).to_ndarray(), filename,
                               color_bounds=color_bounds, cmap_name=cmap_name,
                               func=func)
        else:
            return write_image(self.swapaxes(0, 1)[:, :, channel].to_ndarray(),
                               filename,
                               color_bounds=color_bounds, cmap_name=cmap_name,
                               func=func)

    def save(self, filename, png=True, hdf5=True):
        """
        Saves ImageArray.

        Arguments:
          filename: string
            This should not contain the extension type (.png, .h5, ...)

        Optional Arguments:
          png: boolean, default True
            Save to a png

          hdf5: boolean, default True
            Save to hdf5 file, including info dictionary as attributes.

        """
        if png:
            if len(self.shape) > 2:
                self.write_png("%s.png" % filename)
            else:
                self.write_image("%s.png" % filename)
        if hdf5:
            self.write_hdf5("%s.h5" % filename)



"""
Gizmo frontend tests




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import OrderedDict

from yt.utilities.answer_testing.framework import \
    data_dir_load, \
    requires_ds, \
    sph_answer
from yt.frontends.gizmo.api import GizmoDataset

FIRE_m12i = 'FIRE_M12i_ref11/snapshot_600.hdf5'

# This maps from field names to weight field names to use for projections
fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), ('gas', 'density')),
        (("gas", "metallicity"), ('gas', 'density')),
        (("gas", "O_metallicity"), ('gas', 'density')),
        (('gas', 'velocity_magnitude'), None),
        (("deposit", "all_count"), None),
        (("deposit", "all_cic"), None),
        (("deposit", "PartType0_density"), None),
    ]
)

g64 = "gizmo_64/output/snap_N64L16_135.hdf5"

@requires_ds(g64, big_data=True)
def test_gizmo_64():
    ds = data_dir_load(g64)
    assert isinstance(ds, GizmoDataset)
    for test in sph_answer(ds, 'snap_N64L16_135', 524288, fields):
        test_gizmo_64.__name__ = test.description
        yield test



"""
Test for Composite VR.
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import tempfile
import shutil
from yt.testing import fake_random_ds
from yt.visualization.volume_rendering.api import \
    Scene, \
    VolumeSource, \
    PointSource
import numpy as np
from unittest import TestCase


def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


class PointsVRTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        np.random.seed(0)
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def test_points_vr(self):
        ds = fake_random_ds(64)
        dd = ds.sphere(ds.domain_center, 0.45*ds.domain_width[0])
        ds.field_info[ds.field_list[0]].take_log=False

        sc = Scene()
        cam = sc.add_camera(ds)
        cam.resolution = (512,512)
        vr = VolumeSource(dd, field=ds.field_list[0])
        vr.transfer_function.clear()
        vr.transfer_function.grey_opacity=False
        vr.transfer_function.map_to_colormap(0.0, 1.0, scale=10., colormap="Reds")
        sc.add_source(vr)

        cam.set_width( 1.8*ds.domain_width )
        cam.lens.setup_box_properties(cam)

        # DRAW SOME POINTS
        npoints = 1000
        vertices = np.random.random([npoints, 3])
        colors = np.random.random([npoints, 4])
        colors[:,3] = 0.10

        points_source = PointSource(vertices, colors=colors)
        sc.add_source(points_source)
        im = sc.render()
        im.write_png("points.png")
        return im



"""
A light interface to a few HEALPix routines



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
cimport numpy as np
cimport cython

from libc.stdio cimport fopen, fclose, FILE

cdef extern from "healpix_vectors.h":
    int pix2vec_nest(long nside, long ipix, double *v)
    void vec2pix_nest(long nside, double *vec, long *ipix)
    void pix2ang_nest(long nside, long ipix, double *theta, double *phi)
    void ang2pix_nest(long nside, double theta, double phi, long *ipix)



"""
Halo recipe object



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.operator_registry import \
    OperatorRegistry

recipe_registry = OperatorRegistry()

def add_recipe(name, function):
    recipe_registry[name] =  HaloRecipe(function)

class HaloRecipe(object):
    r"""
    A HaloRecipe is a function that minimally takes in a Halo object
    and performs some analysis on it.  This function may attach attributes
    to the Halo object, write out data, etc, but does not return anything.
    """
    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None: self.args = []
        self.kwargs = kwargs
        if self.kwargs is None: self.kwargs = {}

    def __call__(self, halo_catalog):
        return self.function(halo_catalog, *self.args, **self.kwargs)

def calculate_virial_quantities(hc, fields,
                                weight_field=None, accumulation=True,
                                radius_field="virial_radius", factor=2.0,
                                overdensity_field=("gas", "overdensity"),
                                critical_overdensity=200):
    r"""
    Calculate virial quantities with the following procedure:
    1. Create a sphere data container.
    2. Create 1D radial profiles of overdensity and any requested fields.
    3. Call virial_quantities callback to interpolate profiles for value of critical overdensity.
    4. Delete profile and sphere objects from halo.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    fields: string or list of strings
        The fields for which virial values are to be calculated.
    weight_field : string
        Weight field for profiling.
        Default : "cell_mass"
    accumulation : bool or list of bools
        If True, the profile values for a bin n are the cumulative sum of
        all the values from bin 0 to n.  If -True, the sum is reversed so
        that the value for bin n is the cumulative sum from bin N (total bins)
        to n.  If the profile is 2D or 3D, a list of values can be given to
        control the summation in each dimension independently.
        Default: False.
    radius_field : string
        Field to be retrieved from the quantities dictionary as
        the basis of the halo radius.
        Default: "virial_radius".
    factor : float
        Factor to be multiplied by the base radius for defining
        the radius of the sphere.
        Default: 2.0.
    overdensity_field : string or tuple of strings
        The field used as the overdensity from which interpolation is done to
        calculate virial quantities.
        Default: ("gas", "overdensity")
    critical_overdensity : float
        The value of the overdensity at which to evaulate the virial quantities.
        Overdensity is with respect to the critical density.
        Default: 200

    """

    storage = "virial_quantities_profiles"
    pfields = [field for field in fields if field != "radius"]

    hc.add_callback("sphere", factor=factor)
    if pfields:
        hc.add_callback("profile", ["radius"], pfields,
                        weight_field=weight_field,
                        accumulation=accumulation,
                        storage=storage)
    hc.add_callback("profile", ["radius"], [overdensity_field],
                    weight_field="cell_volume", accumulation=True,
                    storage=storage)
    hc.add_callback("virial_quantities", fields,
                    overdensity_field=overdensity_field,
                    critical_overdensity=critical_overdensity,
                    profile_storage=storage)
    hc.add_callback("delete_attribute", storage)
    hc.add_callback("delete_attribute", "data_object")

add_recipe("calculate_virial_quantities", calculate_virial_quantities)



"""
Definitions for spectral cube coordinate systems




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .cartesian_coordinates import \
    CartesianCoordinateHandler
from .coordinate_handler import \
    _get_coord_fields

class SpectralCubeCoordinateHandler(CartesianCoordinateHandler):
    name = "spectral_cube"

    def __init__(self, ds, ordering = ('x', 'y', 'z')):
        ordering = tuple("xyz"[axis] for axis in
            (ds.lon_axis, ds.lat_axis, ds.spec_axis))
        super(SpectralCubeCoordinateHandler, self).__init__(ds, ordering)

        self.default_unit_label = {}
        names = {}
        if ds.lon_name != "X" or ds.lat_name != "Y":
            names['x'] = "Image\ x"
            names['y'] = "Image\ y"
            # We can just use ds.lon_axis here
            self.default_unit_label[ds.lon_axis] = "pixel"
            self.default_unit_label[ds.lat_axis] = "pixel"
        names['z'] = ds.spec_name
        # Again, can use spec_axis here
        self.default_unit_label[ds.spec_axis] = ds.spec_unit

        self._image_axis_name = ian = {}
        for ax in 'xyz':
            axi = self.axis_id[ax]
            xax = self.axis_name[self.x_axis[ax]]
            yax = self.axis_name[self.y_axis[ax]]
            ian[axi] = ian[ax] = ian[ax.upper()] = (
                names.get(xax, xax), names.get(yax, yax))

        def _spec_axis(ax, x, y):
            p = (x,y)[ax]
            return [self.ds.pixel2spec(pp).v for pp in p]

        self.axis_field = {}
        self.axis_field[self.ds.spec_axis] = _spec_axis

    def setup_fields(self, registry):
        if self.ds.no_cgs_equiv_length is False:
            return super(SpectralCubeCoordinateHandler, self
                    ).setup_fields(registry)
        for axi, ax in enumerate("xyz"):
            f1, f2 = _get_coord_fields(axi)
            def _get_length_func():
                def _length_func(field, data):
                    # Just use axis 0
                    rv = data.ds.arr(data.fcoords[...,0].copy(), field.units)
                    rv[:] = 1.0
                    return rv
                return _length_func
            registry.add_field(("index", "d%s" % ax), sampling_type="cell",  function = f1,
                               display_field = False,
                               units = "code_length")
            registry.add_field(("index", "path_element_%s" % ax), sampling_type="cell", 
                               function = _get_length_func(),
                               display_field = False,
                               units = "")
            registry.add_field(("index", "%s" % ax), sampling_type="cell",  function = f2,
                               display_field = False,
                               units = "code_length")
        def _cell_volume(field, data):
            rv  = data["index", "dx"].copy(order='K')
            rv *= data["index", "dy"]
            rv *= data["index", "dz"]
            return rv
        registry.add_field(("index", "cell_volume"), sampling_type="cell",  function=_cell_volume,
                           display_field=False, units = "code_length**3")
        registry.check_derived_fields(
            [("index", "dx"), ("index", "dy"), ("index", "dz"),
             ("index", "x"), ("index", "y"), ("index", "z"),
             ("index", "cell_volume")])

    _x_pairs = ( ('x', 'y'), ('y', 'x'), ('z', 'x') )
    _y_pairs = ( ('x', 'z'), ('y', 'z'), ('z', 'y') )

    def convert_to_cylindrical(self, coord):
        raise NotImplementedError

    def convert_from_cylindrical(self, coord):
        raise NotImplementedError

    @property
    def image_axis_name(self):
        return self._image_axis_name




import yt
import numpy as np

OCT_MASK_LIST = [8, 0, 0, 0, 0, 8, 0, 0,
                 0, 0, 0, 0, 0, 0, 0, 0,
                 8, 0, 0, 0, 0, 0, 0, 0,
                 0]


def test_octree():
    # See Issue #1272
    octree_mask = np.array(OCT_MASK_LIST, dtype=np.uint8)

    quantities = {}
    quantities[('gas', 'density')] = np.ones((22, 1), dtype=float)

    bbox = np.array([[-10., 10.], [-10., 10.], [-10., 10.]])

    ds = yt.load_octree(octree_mask=octree_mask,
                        data=quantities,
                        bbox=bbox,
                        over_refine_factor=0,
                        partial_coverage=0)

    proj = ds.proj('density', 'x')
    proj['density']



"""
Equivalencies between different kinds of units

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.units.dimensions import temperature, mass, energy, length, rate, \
    velocity, dimensionless, density, number_density, flux, current_cgs, \
    current_mks, charge_cgs, charge_mks, magnetic_field_cgs, magnetic_field_mks, \
    electric_potential_cgs, electric_potential_mks, electric_field_cgs, \
    electric_field_mks, resistance_cgs, resistance_mks

from yt.utilities.physical_ratios import speed_of_light_cm_per_s
from yt.extern.six import add_metaclass
import numpy as np

equivalence_registry = {}

class RegisteredEquivalence(type):
    def __init__(cls, name, b, d):
        type.__init__(cls, name, b, d)
        if hasattr(cls, "_type_name") and not cls._skip_add:
            equivalence_registry[cls._type_name] = cls

@add_metaclass(RegisteredEquivalence)
class Equivalence(object):
    _skip_add = False
    _one_way = False
    def __init__(self):
        import yt.utilities.physical_constants as pc
        self.pc = pc

class NumberDensityEquivalence(Equivalence):
    _type_name = "number_density"
    dims = (density,number_density,)

    def convert(self, x, new_dims, mu=0.6):
        if new_dims == number_density:
            return x/(mu*self.pc.mh)
        elif new_dims == density:
            return x*mu*self.pc.mh

    def __str__(self):
        return "number density: density <-> number density"

class ThermalEquivalence(Equivalence):
    _type_name = "thermal"
    dims = (temperature,energy,)

    def convert(self, x, new_dims):
        if new_dims == energy:
            return self.pc.kboltz*x
        elif new_dims == temperature:
            return x/self.pc.kboltz

    def __str__(self):
        return "thermal: temperature <-> energy"

class MassEnergyEquivalence(Equivalence):
    _type_name = "mass_energy"
    dims = (mass,energy,)

    def convert(self, x, new_dims):
        if new_dims == energy:
            return x*self.pc.clight*self.pc.clight
        elif new_dims == mass:
            return x/(self.pc.clight*self.pc.clight)

    def __str__(self):
        return "mass_energy: mass <-> energy"

class SpectralEquivalence(Equivalence):
    _type_name = "spectral"
    dims = (length,rate,energy,)

    def convert(self, x, new_dims):
        if new_dims == energy:
            if x.units.dimensions == length:
                nu = self.pc.clight/x
            elif x.units.dimensions == rate:
                nu = x
            return self.pc.hcgs*nu
        elif new_dims == length:
            if x.units.dimensions == rate:
                return self.pc.clight/x
            elif x.units.dimensions == energy:
                return self.pc.hcgs*self.pc.clight/x
        elif new_dims == rate:
            if x.units.dimensions == length:
                return self.pc.clight/x
            elif x.units.dimensions == energy:
                return x/self.pc.hcgs

    def __str__(self):
        return "spectral: length <-> rate <-> energy"

class SoundSpeedEquivalence(Equivalence):
    _type_name = "sound_speed"
    dims = (velocity,temperature,energy,)

    def convert(self, x, new_dims, mu=0.6, gamma=5./3.):
        if new_dims == velocity:
            if x.units.dimensions == temperature:
                kT = self.pc.kboltz*x
            elif x.units.dimensions == energy:
                kT = x
            return np.sqrt(gamma*kT/(mu*self.pc.mh))
        else:
            kT = x*x*mu*self.pc.mh/gamma
            if new_dims == temperature:
                return kT/self.pc.kboltz
            else:
                return kT

    def __str__(self):
        return "sound_speed (ideal gas): velocity <-> temperature <-> energy"

class LorentzEquivalence(Equivalence):
    _type_name = "lorentz"
    dims = (dimensionless,velocity,)

    def convert(self, x, new_dims):
        if new_dims == dimensionless:
            beta = x.in_cgs()/self.pc.clight
            return 1./np.sqrt(1.-beta**2)
        elif new_dims == velocity:
            return self.pc.clight*np.sqrt(1.-1./(x*x))

    def __str__(self):
        return "lorentz: velocity <-> dimensionless"

class SchwarzschildEquivalence(Equivalence):
    _type_name = "schwarzschild"
    dims = (mass,length,)

    def convert(self, x, new_dims):
        if new_dims == length:
            return 2.*self.pc.G*x/(self.pc.clight*self.pc.clight)
        elif new_dims == mass:
            return 0.5*x*self.pc.clight*self.pc.clight/self.pc.G

    def __str__(self):
        return "schwarzschild: mass <-> length"

class ComptonEquivalence(Equivalence):
    _type_name = "compton"
    dims = (mass,length,)

    def convert(self, x, new_dims):
        return self.pc.hcgs/(x*self.pc.clight)

    def __str__(self):
        return "compton: mass <-> length"

class EffectiveTemperature(Equivalence):
    _type_name = "effective_temperature"
    dims = (flux,temperature,)

    def convert(self, x, new_dims):
        if new_dims == flux:
            return self.pc.stefan_boltzmann_constant_cgs*x**4
        elif new_dims == temperature:
            return (x/self.pc.stefan_boltzmann_constant_cgs)**0.25

    def __str__(self):
        return "effective_temperature: flux <-> temperature"

em_conversions = {
    charge_mks:("esu", 0.1*speed_of_light_cm_per_s),
    magnetic_field_mks:("gauss", 1.0e4),
    current_mks:("statA", 0.1*speed_of_light_cm_per_s),
    electric_potential_mks:("statV", 1.0e-8*speed_of_light_cm_per_s),
    resistance_mks:("statohm", 1.0e9/(speed_of_light_cm_per_s**2)),
    charge_cgs:("C", 10.0/speed_of_light_cm_per_s),
    magnetic_field_cgs:("T", 1.0e-4),
    current_cgs:("A", 10.0/speed_of_light_cm_per_s),
    electric_potential_cgs:("V", 1.0e8/speed_of_light_cm_per_s),
    resistance_cgs:("ohm", speed_of_light_cm_per_s**2*1.0e-9),
}

class ElectromagneticSI(Equivalence):
    _type_name = "SI"
    _one_way = True
    dims = (current_cgs, charge_cgs, magnetic_field_cgs,
            electric_field_cgs, electric_potential_cgs,
            resistance_cgs)
    
    def convert(self, x, new_dims):
        old_dims = x.units.dimensions
        new_units, convert_factor = em_conversions[old_dims]
        return x.in_cgs().v*convert_factor, new_units
    
    def __str__(self):
        return "SI: EM CGS unit -> EM SI unit"

class ElectromagneticCGS(Equivalence):
    _type_name = "CGS"
    _one_way = True
    dims = (current_mks, charge_mks, magnetic_field_mks,
            electric_field_mks, electric_potential_mks,
            resistance_mks)

    def convert(self, x, new_dims):
        old_dims = x.units.dimensions
        new_units, convert_factor = em_conversions[old_dims]
        return x.in_mks().v*convert_factor, new_units

    def __str__(self):
        return "CGS: EM SI unit -> EM CGS unit"




"""
This module is very simple.  It imports the configuration
we have written for yt.
Everything will be returned in a global config dictionary ``ytcfg``

"""

from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import warnings
from yt.extern.six.moves import configparser

ytcfg_defaults = dict(
    serialize = 'False',
    onlydeserialize = 'False',
    timefunctions = 'False',
    logfile = 'False',
    coloredlogs = 'False',
    suppressstreamlogging = 'False',
    stdoutStreamLogging = 'False',
    loglevel = '20',
    inline = 'False',
    numthreads = '-1',
    __withintesting = 'False',
    __parallel = 'False',
    __global_parallel_rank = '0',
    __global_parallel_size = '1',
    __topcomm_parallel_rank = '0',
    __topcomm_parallel_size = '1',
    __command_line = 'False',
    storeparameterfiles = 'False',
    parameterfilestore = 'parameter_files.csv',
    maximumstoreddatasets = '500',
    skip_dataset_cache = 'True',
    loadfieldplugins = 'True',
    pluginfilename = 'my_plugins.py',
    parallel_traceback = 'False',
    pasteboard_repo = '',
    reconstruct_index = 'True',
    test_storage_dir = '/does/not/exist',
    test_data_dir = '/does/not/exist',
    requires_ds_strict = 'False',
    enzo_db = '',
    hub_url = 'https://girder.hub.yt/api/v1',
    hub_api_key = '',
    hub_sandbox = '/collection/yt_sandbox/data',
    notebook_password = '',
    answer_testing_tolerance = '3',
    answer_testing_bitwise = 'False',
    gold_standard_filename = 'gold311',
    local_standard_filename = 'local001',
    answer_tests_url = 'http://answers.yt-project.org/{1}_{2}',
    sketchfab_api_key = 'None',
    imagebin_api_key = 'e1977d9195fe39e',
    imagebin_upload_url = 'https://api.imgur.com/3/upload',
    imagebin_delete_url = 'https://api.imgur.com/3/image/{delete_hash}',
    curldrop_upload_url = 'http://use.yt/upload',
    thread_field_detection = 'False',
    ignore_invalid_unit_operation_errors = 'False',
    chunk_size = '1000',
    xray_data_dir = '/does/not/exist',
    supp_data_dir = '/does/not/exist',
    default_colormap = 'arbre',
    ray_tracing_engine = 'embree',
    )

CONFIG_DIR = os.environ.get(
    'XDG_CONFIG_HOME', os.path.join(os.path.expanduser('~'), '.config', 'yt'))
if not os.path.exists(CONFIG_DIR):
    try: 
        os.makedirs(CONFIG_DIR)
    except OSError:
        warnings.warn("unable to create yt config directory")

CURRENT_CONFIG_FILE = os.path.join(CONFIG_DIR, 'ytrc')
_OLD_CONFIG_FILE = os.path.join(os.path.expanduser('~'), '.yt', 'config')

# Here is the upgrade.  We're actually going to parse the file in its entirety
# here.  Then, if it has any of the Forbidden Sections, it will be rewritten
# without them.

if os.path.exists(_OLD_CONFIG_FILE):
    f = open(_OLD_CONFIG_FILE).read()
    if any(header in f for header in ["[lagos]","[raven]","[fido]","[enki]"]):
        print("***********************************************************")
        print("* Upgrading configuration file to new format; saving old. *")
        print("***********************************************************")
        # This is of the old format
        cp = configparser.ConfigParser()
        cp.read(_OLD_CONFIG_FILE)
        # NOTE: To avoid having the 'DEFAULT' section here,
        # we are not passing in ytcfg_defaults to the constructor.
        new_cp = configparser.ConfigParser()
        new_cp.add_section("yt")
        for section in cp.sections():
            for option in cp.options(section):
                # We changed them all to lowercase
                if option.lower() in ytcfg_defaults:
                    new_cp.set("yt", option, cp.get(section, option))
                    print("Setting %s to %s" % (option, cp.get(section, option)))
        open(_OLD_CONFIG_FILE + ".old", "w").write(f)
        new_cp.write(open(_OLD_CONFIG_FILE, "w"))

    msg = (
        "The configuration file {} is deprecated. "
        "Please migrate your config to {} by running: "
        "'yt config migrate'"
    )
    warnings.warn(msg.format(_OLD_CONFIG_FILE, CURRENT_CONFIG_FILE))

if not os.path.exists(CURRENT_CONFIG_FILE):
    cp = configparser.ConfigParser()
    cp.add_section("yt")
    try:
        with open(CURRENT_CONFIG_FILE, 'w') as new_cfg:
            cp.write(new_cfg)
    except IOError:
        warnings.warn("unable to write new config file")

class YTConfigParser(configparser.ConfigParser, object):
    def __setitem__(self, key, val):
        self.set(key[0], key[1], val)

    def __getitem__(self, key):
        self.get(key[0], key[1])

    def get(self, section, option, *args, **kwargs):
        val = super(YTConfigParser, self).get(section, option, *args, **kwargs)
        return os.path.expanduser(os.path.expandvars(val))

ytcfg = YTConfigParser(ytcfg_defaults)
ytcfg.read([_OLD_CONFIG_FILE, CURRENT_CONFIG_FILE, 'yt.cfg'])
if not ytcfg.has_section("yt"):
    ytcfg.add_section("yt")

# Now we have parsed the config file.  Overrides come from the command line.

# This should be implemented at some point.  The idea would be to have a set of
# command line options, fed through an option parser, that would override
# the settings in ytcfg.  *However*, because we want to have the command-line
# scripts work, we'd probably want to have them only be long options, and also
# along the lines of --yt-something=somethingelse.  The command line scripts
# would then not get their options from sys.argv, but instead from this module.



"""
EAGLE data-file handling function




"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.frontends.owls.io import \
    IOHandlerOWLS

class IOHandlerEagleNetwork(IOHandlerOWLS):
    _dataset_type = "eagle_network"



from yt.testing import \
    assert_array_equal, \
    assert_equal
import yt
import numpy as np

def setup_fake_refby():
    refine_by=np.array([5, 1, 1])
    top_grid_dim = [100,  10, 2]
    n1=100
    n2=10
    n3=2

    grid_data = [
        dict(left_edge = [0.0, 0.0, 0.0],
             right_edge = [1.0, np.pi, np.pi*2.],
             level = 0,
             dimensions = np.array([n1, n2, n3])),
        dict(left_edge = [0., 0., 0.],
             right_edge = [0.5, np.pi, np.pi*2.],
             level = 1,
             dimensions = refine_by*[n1/2.0, n2, n3]),
    ]

    for g in grid_data:
        g["density"] = (np.random.random(g["dimensions"].astype("i8")),
                        "g/cm**3")
    bbox = np.array([[0.0, 1.0], [0.0, np.pi], [0.0, np.pi*2]])

    ds = yt.load_amr_grids(grid_data, top_grid_dim,
                           bbox = bbox, geometry='spherical',
                           refine_by=refine_by, length_unit='kpc')
    return ds

def test_refine_by():
    ds = setup_fake_refby()
    dd = ds.all_data()
    # This checks that we always refine_by 1 in dimensions 2 and 3
    dims = ds.domain_dimensions*ds.refine_by**ds.max_level
    for i in range(1, 3):
        # Check the refine_by == 1
        ncoords = np.unique(dd.icoords[:,i]).size
        assert_equal(ncoords, dims[i])
    for g in ds.index.grids:
        dims = ds.domain_dimensions*ds.refine_by**g.Level
        # Now we can check converting back to the reference space
        v = ((g.icoords + 1) / dims.astype("f8")).max(axis=0)
        v *= ds.domain_width
        assert_array_equal(v, g.RightEdge.d)



"""
Unit system class.

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.extern.six import string_types
from yt.units import dimensions
from yt.units.unit_object import Unit, unit_system_registry, _get_system_unit_string
from yt.utilities import physical_constants as pc

class UnitSystemConstants(object):
    """
    A class to facilitate conversions of physical constants into a given unit
    system specified by *name*.
    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Physical constants in %s units." % self.name

    def __str__(self):
        return self.name

    def __getattr__(self, item):
        return getattr(pc, item).in_base(self.name)

class UnitSystem(object):
    """
    Create a UnitSystem for facilitating conversions to a default set of units.

    Parameters
    ----------
    name : string
        The name of the unit system. Will be used as the key in the 
        *unit_system_registry* dict to reference the unit system by.
    length_unit : string
        The base length unit of this unit system.
    mass_unit : string
        The base mass unit of this unit system.
    time_unit : string
        The base time unit of this unit system.
    temperature_unit : string, optional
        The base temperature unit of this unit system. Defaults to "K".
    angle_unit : string, optional
        The base angle unit of this unit system. Defaults to "rad".
    current_mks_unit : string, optional
        The base current unit of this unit system. Only used in MKS 
        or MKS-based unit systems.
    registry : :class:`yt.units.unit_registry.UnitRegistry` object
        The unit registry associated with this unit system. Only 
        useful for defining unit systems based on code units.
    """
    def __init__(self, name, length_unit, mass_unit, time_unit,
                 temperature_unit="K", angle_unit="rad", current_mks_unit=None,
                 registry=None):
        self.registry = registry
        self.units_map = {dimensions.length: Unit(length_unit, registry=self.registry),
                          dimensions.mass: Unit(mass_unit, registry=self.registry),
                          dimensions.time: Unit(time_unit, registry=self.registry),
                          dimensions.temperature: Unit(temperature_unit, registry=self.registry),
                          dimensions.angle: Unit(angle_unit, registry=self.registry)}
        self._dims = ["length","mass","time","temperature","angle"]
        if current_mks_unit is not None:
            self.units_map[dimensions.current_mks] = Unit(current_mks_unit, registry=self.registry)
            self._dims.append("current_mks")
        self.registry = registry
        self.base_units = self.units_map.copy()
        unit_system_registry[name] = self
        self.name = name
        self.constants = UnitSystemConstants(self.name)

    def __getitem__(self, key):
        if isinstance(key, string_types):
            key = getattr(dimensions, key)
        um = self.units_map
        if key not in um or um[key].dimensions is not key:
            units = _get_system_unit_string(key, self.units_map)
            self.units_map[key] = Unit(units, registry=self.registry)
        return self.units_map[key]

    def __setitem__(self, key, value):
        if isinstance(key, string_types):
            if key not in self._dims:
                self._dims.append(key)
            key = getattr(dimensions, key)
        self.units_map[key] = Unit(value, registry=self.registry)

    def __str__(self):
        return self.name

    def __repr__(self):
        repr = "%s Unit System\n" % self.name
        repr += " Base Units:\n"
        for dim in self.base_units:
            repr += "  %s: %s\n" % (str(dim).strip("()"), self.base_units[dim])
        repr += " Other Units:\n"
        for key in self._dims:
            dim = getattr(dimensions, key)
            if dim not in self.base_units:
                repr += "  %s: %s\n" % (key, self.units_map[dim])
        return repr

def create_code_unit_system(unit_registry, current_mks_unit=None):
    code_unit_system = UnitSystem(unit_registry.unit_system_id, "code_length", 
                                  "code_mass", "code_time", "code_temperature",
                                  current_mks_unit=current_mks_unit, 
                                  registry=unit_registry)
    code_unit_system["velocity"] = "code_velocity"
    if current_mks_unit:
        code_unit_system["magnetic_field_mks"] = "code_magnetic"
    else:
        code_unit_system["magnetic_field_cgs"] = "code_magnetic"
    code_unit_system["pressure"] = "code_pressure"

cgs_unit_system = UnitSystem("cgs", "cm", "g", "s")
cgs_unit_system["energy"] = "erg"
cgs_unit_system["specific_energy"] = "erg/g"
cgs_unit_system["pressure"] = "dyne/cm**2"
cgs_unit_system["force"] = "dyne"
cgs_unit_system["magnetic_field_cgs"] = "gauss"
cgs_unit_system["charge_cgs"] = "esu"
cgs_unit_system["current_cgs"] = "statA"

mks_unit_system = UnitSystem("mks", "m", "kg", "s", current_mks_unit="A")
mks_unit_system["energy"] = "J"
mks_unit_system["specific_energy"] = "J/kg"
mks_unit_system["pressure"] = "Pa"
mks_unit_system["force"] = "N"
mks_unit_system["magnetic_field_mks"] = "T"
mks_unit_system["charge_mks"] = "C"

imperial_unit_system = UnitSystem("imperial", "ft", "lbm", "s", 
                                  temperature_unit="R")
imperial_unit_system["force"] = "lbf"
imperial_unit_system["energy"] = "ft*lbf"
imperial_unit_system["pressure"] = "lbf/ft**2"

galactic_unit_system = UnitSystem("galactic", "kpc", "Msun", "Myr")
galactic_unit_system["energy"] = "keV"
galactic_unit_system["magnetic_field_cgs"] = "uG"

solar_unit_system = UnitSystem("solar", "AU", "Mearth", "yr")

geometrized_unit_system = UnitSystem("geometrized", "l_geom", 
                                     "m_geom", "t_geom")

planck_unit_system = UnitSystem("planck", "l_pl", "m_pl", "t_pl", 
                                temperature_unit="T_pl")
planck_unit_system["energy"] = "E_pl"
planck_unit_system["charge_cgs"] = "q_pl"


cgs_ampere_unit_system = UnitSystem('cgs-ampere', 'cm', 'g', 's',
                                    current_mks_unit='A')
cgs_ampere_unit_system["energy"] = "erg"
cgs_ampere_unit_system["specific_energy"] = "erg/g"
cgs_ampere_unit_system["pressure"] = "dyne/cm**2"
cgs_ampere_unit_system["force"] = "dyne"
cgs_ampere_unit_system["magnetic_field_cgs"] = "gauss"
cgs_ampere_unit_system["charge_cgs"] = "esu"
cgs_ampere_unit_system["current_cgs"] = "statA"



from yt.units.unit_object import define_unit
from yt.units.yt_array import YTQuantity
from yt.convenience import load
from yt.testing import requires_file

def test_define_unit():
    define_unit("mph", (1.0, "mile/hr"))
    a = YTQuantity(2.0, "mph")
    b = YTQuantity(1.0, "mile")
    c = YTQuantity(1.0, "hr")
    assert a == 2.0*b/c
    d = YTQuantity(1000.0, "cm**3")
    define_unit("L", d, prefixable=True)
    e = YTQuantity(1.0, "mL")
    f = YTQuantity(1.0, "cm**3")
    assert e == f

gslr = "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0300"
@requires_file(gslr)
def test_define_unit_dataset():
    ds = load(gslr)
    ds.define_unit("fortnight", (14.0, "day"))
    a = ds.quan(1.0, "fortnight")
    b = ds.quan(3600.0*24.0*14.0, "code_time")
    assert a == b
